"use strict";
// Copyright (c) Mito
// Distributed under the terms of the Modified BSD License.
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
/*
  Contains a wrapper around a strongly typed object that simulates a web-api.
*/
// Max delay is the longest we'll wait for the API to return a value
const MAX_DELAY = 5000;
// How often we poll to see if we have a response yet
const RETRY_DELAY = 250;
const MAX_RETRIES = MAX_DELAY / RETRY_DELAY;
const getRandomId = () => {
    return '_' + Math.random().toString(36).substr(2, 9);
};
/*
  The MitoAPI class contains functions for interacting with the Mito backend. All
  interactions with the backend should go through this call, so that if we ever move
  to a more standard API, the migration is protected by this interface!
*/
class MitoAPI {
    constructor(send) {
        this.send = send;
        this.unconsumedResponses = [];
    }
    /*
        The receive response function is a workaround to the fact that we _do not_ have
        a real API in practice. If/when we do have a real API, we'll get rid of this function,
        and allow the API to just make a call to a server, and wait on a response
    */
    receiveResponse(response) {
        this.unconsumedResponses.push(response);
    }
    /*
        Helper function that tries to get the response for a given ID, and returns
        the data inside the 'data' key in this response if it exists.

        Returns undefined if it does not get a response within the set timeframe
        for retries.
    */
    getResponseData(id) {
        return new Promise((resolve) => {
            let tries = 0;
            const interval = setInterval(() => {
                // Only try at most MAX_RETRIES times
                tries++;
                if (tries > MAX_RETRIES) {
                    clearInterval(interval);
                    // If we fail, we return an empty response
                    return resolve(undefined);
                }
                // See if there is an API response to this one specificially
                const index = this.unconsumedResponses.findIndex((response) => {
                    return response['id'] === id;
                });
                if (index !== -1) {
                    // Clear the interval
                    clearInterval(interval);
                    const response = this.unconsumedResponses[index];
                    this.unconsumedResponses.splice(index, 1);
                    return resolve(response['data']); // return to end execution
                }
                else {
                    console.log("Still waiting");
                }
            }, RETRY_DELAY);
        });
    }
    /*
        Returns all the CSV files in the current folder as the kernel.
    */
    getDataFiles() {
        return __awaiter(this, void 0, void 0, function* () {
            const id = getRandomId();
            this.send({
                'event': 'api_call',
                'type': 'datafiles',
                'id': id
            });
            const dataFiles = yield this.getResponseData(id);
            if (dataFiles == undefined) {
                return [];
            }
            return dataFiles;
        });
    }
    /*
        Import summaries are a mapping from step_idx -> import information for each of the
        import steps in the analysis with the given analysisName.
    */
    getImportSummary(analysisName) {
        return __awaiter(this, void 0, void 0, function* () {
            const id = getRandomId();
            this.send({
                'event': 'api_call',
                'type': 'import_summary',
                'id': id,
                'analysis_name': analysisName
            });
            const importSumary = yield this.getResponseData(id);
            if (importSumary == undefined) {
                return {};
            }
            return importSumary;
        });
    }
    /*
        Adds a column with the passed parameters
    */
    sendColumnAddMessage(sheetIndex, columnHeader, stepID) {
        return __awaiter(this, void 0, void 0, function* () {
            if (stepID === undefined || stepID == '') {
                stepID = getRandomId();
            }
            this.send({
                'event': 'edit_event',
                'type': 'add_column_edit',
                'sheet_index': sheetIndex,
                'column_header': columnHeader
            });
            return stepID;
        });
    }
    /*
        Adds a delete column message with the passed parameters
    */
    sendDeleteColumn(sheetIndex, columnHeader) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'edit_event',
                'type': 'delete_column_edit',
                'sheet_index': sheetIndex,
                'column_header': columnHeader
            });
        });
    }
    /*
        Does a merge with the passed parameters, returning the ID of the edit
        event that was generated (in case you want to overwrite it).
    */
    sendMergeMessage(sheetOneIndex, sheetOneMergeKey, sheetOneSelectedColumns, sheetTwoIndex, sheetTwoMergeKey, sheetTwoSelectedColumns, 
    /*
        If you want to overwrite, you have to pass the ID of the the step that
        you want to overwrite. Not passing this argument, or passing an empty string,
        will result in no overwrite occuring (and a new stepID) being returned.
    */
    stepID) {
        return __awaiter(this, void 0, void 0, function* () {
            // If this is overwriting a merge event, then we do not need to
            // create a new id, as we already have it!
            if (stepID === undefined || stepID == '') {
                stepID = getRandomId();
            }
            this.send({
                event: 'edit_event',
                type: 'merge_edit',
                step_id: stepID,
                sheet_index_one: sheetOneIndex,
                merge_key_one: sheetOneMergeKey,
                selected_columns_one: sheetOneSelectedColumns,
                sheet_index_two: sheetTwoIndex,
                merge_key_two: sheetTwoMergeKey,
                selected_columns_two: sheetTwoSelectedColumns,
            });
            return stepID;
        });
    }
    /*
        Does a pivot with the passed parameters, returning the ID of the edit
        event that was generated (in case you want to overwrite it).
    */
    sendPivotMessage(sheetIndex, pivotRows, pivotCols, values, stepID) {
        return __awaiter(this, void 0, void 0, function* () {
            // If this is overwriting a pivot event, then we do not need to
            // create a new id, as we already have it!
            if (stepID === undefined || stepID === '') {
                stepID = getRandomId();
            }
            this.send({
                event: 'edit_event',
                type: 'pivot_edit',
                sheet_index: sheetIndex,
                pivot_rows: pivotRows,
                pivot_columns: pivotCols,
                values: values,
                step_id: stepID
            });
            return stepID;
        });
    }
    /*
        Reorders the columnHeader on sheetIndex to the newIndex, and shifts the remaining
        columns to the right.
    */
    sendReorderColumnMessage(sheetIndex, columnHeader, newIndex) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'edit_event',
                'type': 'reorder_column_edit',
                'sheet_index': sheetIndex,
                'column_header': columnHeader,
                'new_column_index': newIndex
            });
        });
    }
    /*
        Renames the dataframe at sheetIndex.
    */
    sendDataframeRenameEdit(sheetIndex, newDataframeName, stepID) {
        return __awaiter(this, void 0, void 0, function* () {
            // If this is overwriting a pivot event, then we do not need to
            // create a new id, as we already have it!
            if (stepID === undefined || stepID === '') {
                stepID = getRandomId();
            }
            this.send({
                'event': 'edit_event',
                'type': 'dataframe_rename_edit',
                'sheet_index': sheetIndex,
                'new_dataframe_name': newDataframeName
            });
            return stepID;
        });
    }
    /*
        Does a filter with the passed parameters, returning the ID of the edit
        event that was generated (in case you want to overwrite it).
    */
    sendFilterMessage(sheetIndex, columnHeader, filters, operator, stepID) {
        return __awaiter(this, void 0, void 0, function* () {
            // create a new id, if we need it!
            if (stepID === undefined || stepID === '') {
                stepID = getRandomId();
            }
            this.send({
                event: 'edit_event',
                type: 'filter_column_edit',
                sheet_index: sheetIndex,
                column_header: columnHeader,
                operator: operator,
                filters: filters,
                step_id: stepID
            });
            return stepID;
        });
    }
    /*
        Does a sort with the passed parameters, returning the ID of the edit
        event that was generated (in case you want to overwrite it).
    */
    sendSortMessage(sheetIndex, columnHeader, sortDirection, stepID) {
        return __awaiter(this, void 0, void 0, function* () {
            if (stepID === undefined || stepID === '') {
                stepID = getRandomId();
            }
            this.send({
                event: 'edit_event',
                type: 'sort_edit',
                sheet_index: sheetIndex,
                column_header: columnHeader,
                sort_direction: sortDirection,
                step_id: stepID
            });
            return stepID;
        });
    }
    /*
        Renames a column with the passed parameters
    */
    sendRenameColumn(sheetIndex, oldColumnHeader, newColumnHeader, stepID) {
        return __awaiter(this, void 0, void 0, function* () {
            if (stepID === undefined || stepID === '') {
                stepID = getRandomId();
            }
            this.send({
                event: 'edit_event',
                type: 'rename_column_edit',
                sheet_index: sheetIndex,
                old_column_header: oldColumnHeader,
                new_column_header: newColumnHeader,
                stepID: stepID
            });
            return stepID;
        });
    }
    /*
        Duplicates the dataframe at sheetIndex.
    */
    sendDataframeDuplicateMessage(sheetIndex) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'edit_event',
                'type': 'dataframe_duplicate_edit',
                'sheet_index': sheetIndex,
            });
        });
    }
    /*
        Deletes the dataframe at the passed sheetIndex
    */
    sendDataframeDeleteMessage(sheetIndex) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'edit_event',
                'type': 'dataframe_delete_edit',
                'sheet_index': sheetIndex,
            });
        });
    }
    /*
        Sends an undo message, which removes the last step that was created.
    */
    sendUndoMessage() {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'update_event',
                'type': 'undo'
            });
        });
    }
    /*
        Sends a message which saves the current analysis with the passed
        analysisName.
    */
    sendSaveAnalysisMessage(analysisName) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'update_event',
                'type': 'save_analysis_update',
                'analysis_name': analysisName
            });
        });
    }
    /*
        Sends a message to tell Mito to replay an existing analysis onto
        the current analysis.
    */
    sendUseExistingAnalysisUpdateMessage(analysisName, newFileNames, clearExistingAnalysis) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'update_event',
                'type': 'replay_analysis_update',
                'analysis_name': analysisName,
                'import_summaries': newFileNames === undefined ? {} : newFileNames,
                'clear_existing_analysis': clearExistingAnalysis === undefined ? false : clearExistingAnalysis
            });
        });
    }
    /*
        Sets the formula for the given columns.
    */
    sendSetColumnFormulaEditMessage(sheetIndex, columnHeader, newFormula) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'edit_event',
                'type': 'set_column_formula_edit',
                'sheet_index': sheetIndex,
                'column_header': columnHeader,
                'new_formula': newFormula
            });
        });
    }
    /*
        Imports the given file names.
    */
    sendSimpleImportMessage(fileNames) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'edit_event',
                'type': 'simple_import_edit',
                'file_names': fileNames,
            });
        });
    }
    /*
        Does an import by running the pythonCode, and pulling out
        the new df names.
    */
    sendRawPythonImportMessage(pythonCode, newDfNames) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'edit_event',
                'type': 'raw_python_import_edit',
                'python_code': pythonCode,
                'new_df_names': newDfNames,
            });
        });
    }
    /*
        Sends the user_email to the backend so the user can sign in
    */
    sendSignUp(userEmail) {
        return __awaiter(this, void 0, void 0, function* () {
            this.send({
                'event': 'update_event',
                'type': 'signup_update',
                'user_email': userEmail
            });
        });
    }
    /*
        Sends a log event from the frontend to the backend, where it is logged
        by the backend. We log in the backend to keep a linear stream of actions
        that is making.
    */
    sendLogMessage(logEventType, params) {
        return __awaiter(this, void 0, void 0, function* () {
            let message = {};
            // Copy the params, so we don't accidently modify anything
            if (params !== undefined) {
                message = Object.assign({}, params);
            }
            // Save the type of event, as well as what is being logged
            message['event'] = 'log_event';
            message['type'] = logEventType;
            this.send(message);
        });
    }
}
exports.MitoAPI = MitoAPI;
//# sourceMappingURL=api.js.map